#ifndef F_CPU
	#define F_CPU 4000000UL
#endif
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>


#define KEY_DDR         DDRB
#define KEY_PORT        PORTB
#define KEY_PIN         PINB
#define KEY0            PB0
#define KEY1            PB1
#define ALL_KEYS        (1<<KEY0 | 1<<KEY1)
 
#define REPEAT_MASK     (1<<KEY0 | 1<<KEY1)       // repeat: key1, key2
#define REPEAT_START    50                        // after 500ms
#define REPEAT_NEXT     20                        // every 200ms


volatile uint8_t key_state;                                // debounced and inverted key state:
                                                  // bit = 1: key pressed
volatile uint8_t key_press;                                // key press detect

volatile uint8_t key_rpt;                                  // key long press and repeatedly

volatile uint8_t zero_cnt = 0;
volatile uint8_t once = 0;
volatile uint8_t duty = 0;

ISR( TIMER0_OVF_vect )                            // every 10ms
{
  static uint8_t ct0, ct1, rpt;
  uint8_t i;
 
  TCNT0 = (uint8_t)(int16_t)-(F_CPU / 1024 * 10e-3 + 0.5);  // preload for 10ms
 
  i = key_state ^ ~KEY_PIN;                       // key changed ?
  ct0 = ~( ct0 & i );                             // reset or count ct0
  ct1 = ct0 ^ (ct1 & i);                          // reset or count ct1
  i &= ct0 & ct1;                                 // count until roll over ?
  key_state ^= i;                                 // then toggle debounced state
  key_press |= key_state & i;                     // 0->1: key press detect
 
  if( (key_state & REPEAT_MASK) == 0 )            // check repeat function
     rpt = REPEAT_START;                          // start delay
  if( --rpt == 0 ){
    rpt = REPEAT_NEXT;                            // repeat delay
    key_rpt |= key_state & REPEAT_MASK;
  }
}

ISR( INT0_vect ) {
	if( ++zero_cnt >= 20 ) {
		// Ab 20. Halbwelle TRIAC wieder loeschen (ausser auf Stufe 3)
		zero_cnt = 0;
		if( duty != 3 ) {
			once = 0;
			PORTD &= ~( 1<<PD0 );
		}
	}
}


void IOinit( void ) {
	// PD0 und PD1 als Ausgang
	// PD0: Triac fuer Motor
	// PD3,4,5: LEDs fuer Geschwindigkeitsanzeige
	// PD6: Licht
	DDRD = ( 1<<PD0 ) | ( 1<<PD3 ) | ( 1<<PD4 ) | ( 1<<PD5 ) | ( 1<<PD6 );

	// Alle Ausgänge ausschalten
	PORTD &= ~( 1<<PD0 ) & ~( 1<<PD3 ) & ~( 1<<PD4 ) & ~( 1<<PD5 ) & ~( 1<<PD6 );

	// PB0 ist Taster fuer Luefter
	// PB1 ist Taster fuer Licht
	DDRB = 0;
	// Interner Pullup aktivieren
	PORTB |= (1<<PB0) | (1<<PD1);

	
	// PD2 - zero-crossing detector
	PCMSK |= (1<<PD2);
	// interrupt on INT0 pin falling edge (sensor triggered) 
	MCUCR = (1<<ISC01) | (1<<ISC00);
	// turn on interrupts!
	GIMSK  |= (1<<INT0);
	
}

///////////////////////////////////////////////////////////////////
//
// check if a key has been pressed. Each pressed key is reported
// only once
//
uint8_t get_key_press( uint8_t key_mask )
{
  cli();                                          // read and clear atomic !
  key_mask &= key_press;                          // read key(s)
  key_press ^= key_mask;                          // clear key(s)
  sei();
  return key_mask;
}
 
///////////////////////////////////////////////////////////////////
//
// check if a key has been pressed long enough such that the
// key repeat functionality kicks in. After a small setup delay
// the key is reported being pressed in subsequent calls
// to this function. This simulates the user repeatedly
// pressing and releasing the key.
//
uint8_t get_key_rpt( uint8_t key_mask )
{
  cli();                                          // read and clear atomic !
  key_mask &= key_rpt;                            // read key(s)
  key_rpt ^= key_mask;                            // clear key(s)
  sei();
  return key_mask;
}
 
///////////////////////////////////////////////////////////////////
//
// check if a key is pressed right now
//
uint8_t get_key_state( uint8_t key_mask )
 
{
  key_mask &= key_state;
  return key_mask;
}
 
///////////////////////////////////////////////////////////////////
//
uint8_t get_key_short( uint8_t key_mask )
{
  cli();                                          // read key state and key press atomic !
  return get_key_press( ~key_state & key_mask );
}
 
///////////////////////////////////////////////////////////////////
//
uint8_t get_key_long( uint8_t key_mask )
{
  return get_key_press( get_key_rpt( key_mask ));
}

void pause( uint16_t ms ) {
	/* pauses the program for the given time;
	* ms: time in milliseconds (0 - 65535)
	*/

	uint16_t i;

	for( i = 0; i < ms / 63; i++ ) {
		_delay_ms( 63 );
	}

	_delay_ms( ms % 63 );
}

int main( void ) {
	IOinit();
	
	// Configure debouncing routines
	KEY_DDR &= ~ALL_KEYS;                // configure key port for input
	KEY_PORT |= ALL_KEYS;                // and turn on pull up resistors

	TCCR0B = (1<<CS02)|(1<<CS00);         // divide by 1024
	TCNT0 = (uint8_t)(int16_t)-( F_CPU / 1024 * 10e-3 + 0.5 );  // preload for 10ms
	TIMSK |= 1<<TOIE0;                   // enable timer interrupts

	sei();
	
	
	while( 1 ) {
		if( get_key_short( 1<<KEY0 )) {
			if( ++duty > 3 ) {
				duty = 0;
			}
			switch( duty ) {
				case 0:
					// LEDs ausschalten
					PORTD &= ~( 1<<PD3 ) & ~( 1<<PD4 ) & ~( 1<<PD5 );
					break;
				case 1:
					// LED 1 an
					PORTD |= ( 1<<PD3 );
					break;
				case 2:
					// LED 1 und 2 an
					PORTD |= ( 1<<PD3 ) | ( 1<<PD4 );
					break;
				case 3:
					// LED 1, 2 und 3 an
					PORTD |= ( 1<<PD3 ) | ( 1<<PD4 ) | ( 1<<PD5 );
					break;
			}
		}
		if( get_key_long( 1<<KEY0 )) {
			// Lang druecken zum Ausschalten...
			duty = 0;
			PORTD &= ~( 1<<PD3 ) & ~( 1<<PD4 ) & ~( 1<<PD5 );
		}

		// Licht-Schalter
		if( get_key_short( 1<<KEY1 )) {
			PORTD ^= ( 1 << PD6 );
		}
		
		switch( duty ) {
			// Auf Stufe 1 ab der 12. Halbwelle bis zur 20. schalten
			case 1:
				if( zero_cnt > 12 && once == 0 ) {
					PORTD |= ( 1<<PD0 );
					once = 1;
				}
				break;
			// Auf Stufe 2 ab der 7. Halbwelle bis zur 20. schalten
			case 2:
				if( zero_cnt > 7 && once == 0 ) {
					PORTD |= ( 1<<PD0 );
					once = 1;
				}
				break;
			// Auf Stufe 3 komplett durchschalten
			case 3:
				if( once == 0 ) {
					PORTD |= ( 1<<PD0 );
					once = 1;
				}
				break;
		}
	}
	return( 0 );
}
